/*
 * Fits.c
 *
 * by John Pavan
 *
 * Functions for use with gnuscope
 * which are similar to fits or
 * help when using fits
 */

#include <gtk/gtk.h>
#include "gtkgraph.h"
#include <math.h>
#include <stdio.h>
#include "fits.h"
#include "gnuscopefuncs.h"
#include "gnuscopeglobals.h"

/* --- Globals --- */
//int histid[254],histsize[254],histpointer;
//int *histloc[254];
//char field1[254][40],field2[254][40],field3[254][40];
int intype,spectra,plottype;
GtkWidget *graph;
int markers[4],binsize,offset,currentrange[2];
float scaling;
int yscale,ybin,yscaleforce,ybinforce;
long double displayused;
int background;
//int globalcalibration[3];//,calibration[254][3];
int numspectra;
GtkWidget *box1,*graphsdisplayed[16];
int displaygaussian,displaydoublegaussian;
GtkWidget *maintext;
struct stoppingpower *stoppingpowerinfo;
GtkWidget *population[10][4];
GtkWidget *tauentry[10][4];
GtkWidget *tsfentry[10][4];
struct fitsvariables *fitsstuff;

/* --- Function Declarations --- */

void GetMessageDialog(const char *message);
void WriteMainText(const char *newtext);

void FitsReadStoppingPower(char *sFilename);
GtkWidget *MakeFeedingEntry(int level, int pos);
void FeedingFileSaveCallback(GtkWidget *widget, gpointer *data);
void FeedingFileSave(char *sFilename);
void FeedingFileReadCallback(GtkWidget *widget, gpointer *data);
void FeedingFileRead(char *sFilename);
void FeedingClear(GtkWidget *widget, gpointer *data);
void ReadFitsSetup(char *sFilename);
void Fits();
int CompareInts(const void *a, const void *b);
void FitsTauPrompt(GtkWidget *text);
void DepopulateLevels(float time, int flag);
void Shift(float enw, int jj);
void FitsTauEntry(GtkWidget *widget, GtkWidget *entry);
void FitsAvCos(float a1, float z1, float a2, float z2, float er);
void DetectorGeometry();
float AP2(float x);
float AP4(float x);
float CrossSection(float beamenergy);
float LinearInterPol(float x1, float x2, float y1, float y2, float x);
void Kinemat(float beamenergy, float recoilenergy);
void GetStoppingPower(float ei, int k, float *ef, int kstop, float *dedx);
void ChiSqr();

float ProjectileStop(float beamenergy);

/* --- functions --- */

/* ListBetweenMarkers
 *
 * lists the counts in the channels between (and including) the markers
 */
void ListBetweenMarkers()
{
  int i;
  int localmarkers[2];
  char dummystr[80];

  /* --- get the markers in order --- */
  if (markers[0] > markers[1]) {
    localmarkers[0] = markers[1];
    localmarkers[1] = markers[0];
  } else {
    localmarkers[1] = markers[1];
    localmarkers[0] = markers[0];
  }

  for (i = localmarkers[1] ; i >= localmarkers[0];i--) {
    sprintf(dummystr,"%d ",*(histloc[spectra] + i));
    WriteMainText(dummystr);
    if ((i % 10) == 0) {
      WriteMainText("\n");
    }
  }
  sprintf(dummystr,"Counts between channels %d and %d.\n",(localmarkers[0]+1),(localmarkers[1]+1));
  WriteMainText(dummystr);
}

/* FitsReadStoppingPower
 *
 * Reads in a Stopping Power file of the same format as fits.
 * Details of what is being read in will be included in
 * comments of the code.
 */
void FitsReadStoppingPower(char *sFilename)
{
  int i,j,k,l,m;
  FILE *infile;
  char dummystr[160];
  float a,b,c,d,e,f,g,h;
  char *charpointer;

  if ((infile = fopen(sFilename,"r")) != NULL) {
    /* --- first if stoppingpowerinfo has not already been mallocated, we need to
       --- mallocate it --- */
    if (stoppingpowerinfo == NULL) {
      stoppingpowerinfo = (struct stoppingpower *) malloc(sizeof(struct stoppingpower));
      stoppingpowerinfo->rectheta = NULL;
      stoppingpowerinfo->recphi = NULL;
      stoppingpowerinfo->recprobtheta = NULL;
      stoppingpowerinfo->projstop = NULL;
      stoppingpowerinfo->stopen1 = NULL;
      stoppingpowerinfo->stop1 = NULL;
      stoppingpowerinfo->stopen2 = NULL;
      stoppingpowerinfo->stop2 = NULL;
      stoppingpowerinfo->crosssectionen = NULL;
      stoppingpowerinfo->crosssection = NULL;
    }
    /* --- skip the first line, we won't be using it --- */
    /* --- the first line is a comment about what the file contains --- */
    fgets(dummystr,80,infile);
    /* --- now read in the projectile mass and energy --- */
    if (fscanf(infile,"%f,%f",&a,&b) == 2) {
      stoppingpowerinfo->projmass = a;
      stoppingpowerinfo->beamenergy = b;
    } else {
      GetMessageDialog("Error Reading the projectile mass and beam energy.\n");
      goto stopreading;
    }
    /* --- get the target a,z and the backing a,z --- */
    if (fscanf(infile,"%d,%d,%d,%d",&i,&j,&k,&l) == 4) {
      stoppingpowerinfo->targa = i;
      stoppingpowerinfo->targz = j;
      stoppingpowerinfo->backa = k;
      stoppingpowerinfo->backz = l;
    } else {
      GetMessageDialog("Error Reading the target A,Z and the backing A,Z.\n");
      goto stopreading;
    }
    /* --- now for:
       --- blaugrund's approx switch
       --- fractional FWHM of recoil energy distribution
       --- number of recoil theta angles
       --- number of recoil phi angles (why?  They arn't used)
       --- */
    if (fscanf(infile,"%d,%d,%d,%f,%d,%d",&i,&j,&k,&a,&l,&m) == 6) {
      stoppingpowerinfo->reca = i;
      stoppingpowerinfo->recz = j;
      stoppingpowerinfo->blauswitch = k;
      stoppingpowerinfo->recfwhmen = a;
      stoppingpowerinfo->numrectheta = l;
      stoppingpowerinfo->numrecphi = m;
    } else {
      GetMessageDialog("Error Reading recoil information.\n");
      goto stopreading;
    }
    /* --- now we need to read in the recoil thetas and probablilities --- */
    /* --- first get the right amount of memory --- */
    if (stoppingpowerinfo->rectheta == NULL) {
      stoppingpowerinfo->rectheta = (float *)
	malloc(stoppingpowerinfo->numrectheta * sizeof(float));
    } else {
      stoppingpowerinfo->rectheta = (float *)
	realloc(stoppingpowerinfo->rectheta, stoppingpowerinfo->numrectheta * sizeof(float));
    }
    if (stoppingpowerinfo->recprobtheta == NULL) {
      stoppingpowerinfo->recprobtheta = (float *) 
	malloc(stoppingpowerinfo->numrectheta * sizeof(float));
    } else {
      stoppingpowerinfo->recprobtheta = (float *)
	realloc(stoppingpowerinfo->recprobtheta, stoppingpowerinfo->numrectheta * sizeof(float));
    }
    /* --- now let's read the info --- */
    for (i = 0; i < stoppingpowerinfo->numrectheta; i++) {
      if (fscanf(infile,"%f,%f",&a,&b) == 2) {
	stoppingpowerinfo->rectheta[i] = a;
	stoppingpowerinfo->recprobtheta[i] = b;
      } else {
	GetMessageDialog("Error reading recoil angles.");
	goto stopreading;
      }
    }
    /* --- now let's read the detector angles/resolutions --- */
    if (fscanf(infile,"%f,%f,%f",&a,&b,&c) == 3) {
      stoppingpowerinfo->detres[0] = a;
      stoppingpowerinfo->detres[1] = b;
      stoppingpowerinfo->detres[2] = c;
    } else {
      GetMessageDialog("Error Reading the detector resolutions.");
      goto stopreading;
    }
    /* --- now we need to read:
       --- the projectile stopping power normalization
       --- the minium beam energy
       --- the energy increment for projectile stopping powers
       --- the number of projectiles stopping powers
       --- */
    if (fscanf(infile,"%f,%f,%f,%d",&a,&b,&c,&i) == 4) {
      stoppingpowerinfo->projstopnorm = a;
      stoppingpowerinfo->minbeamen = b;
      stoppingpowerinfo->debeamen = c;
      stoppingpowerinfo->numprojstop = i;
    } else {
      GetMessageDialog("Error reading projectile information.");
      goto stopreading;
    }
    /* --- now we need to read in the projectile stopping powers --- */
    /* --- first we need to allocate enough memory --- */
    if (stoppingpowerinfo->projstop == NULL) {
      stoppingpowerinfo->projstop = 
	(float *) malloc(stoppingpowerinfo->numprojstop * sizeof(float));
    } else {
      stoppingpowerinfo->projstop = 
	(float *) realloc(stoppingpowerinfo->projstop, 
			  stoppingpowerinfo->numprojstop * sizeof(float));
    }
    /* --- now let's read it in --- */
    /* --- this is going to be interesting because they tend to be written
       --- in one row. --- */
    i = 0;
    //    printf("number of projectile stopping powers %d.\n",
    //	   stoppingpowerinfo->numprojstop);
      fgets(dummystr,160,infile);
    while (i < stoppingpowerinfo->numprojstop) {
      /* --- read in a line --- */
      //printf("about to read line.\n");
      if (fgets(dummystr,160,infile) == NULL) {
	GetMessageDialog("Error reading projectile stopping powers.");
      	goto stopreading;
      }
      //printf("read line.\n");
      /* --- now we need to extract the information from that string --- */
      charpointer = (char *) &dummystr;
      //printf(dummystr);
      //printf(charpointer);
      while ((charpointer != NULL) && (i < stoppingpowerinfo->numprojstop)) {
	if (sscanf(charpointer,"%f",&a) == 1) {
	  stoppingpowerinfo->projstop[i] = a;
	  //printf("%f\n",a);
	  i++;
	  //printf("about to get a new pointer.\n");
	  charpointer = (char *) strstr(charpointer,",");
	  if (charpointer != NULL) {
	    charpointer = &charpointer[1];
	    // printf(charpointer);
	  }
	  //printf("got a new pointer.\n");
	} else {
	  GetMessageDialog("Error reading projectile stopping powers.");
	  goto stopreading;
	}
      }
    }
    /* --- now we need to read in:
       --- the number of target elements,
       --- the element thickness,
       --- the number of coarse stopping powers
       --- the stopping power normalization,
       --- the number of coarse stoping powers to read
       --- and the number of recoils to start with.
       --- */

    if (fscanf(infile,"%d,%f,%d,%f,%d,%f,%d",&i,&a,&k,&b,&l,&c,&m) == 7) {
      stoppingpowerinfo->numtarel = i;
      stoppingpowerinfo->tarelthick = a;
      stoppingpowerinfo->numcoarsestop = k;
      stoppingpowerinfo->tarstopnorm[0] = b;
      stoppingpowerinfo->numrecoils[0] = l;
      stoppingpowerinfo->tarstopnorm[1] = c;
      stoppingpowerinfo->numrecoils[1] = m;
    } else {
      GetMessageDialog("Error Reading the stopping power initialization.\n");
      goto stopreading;
    }
    /* --- now we need to read some floats that I don't know what to do with --- */

    if (fscanf(infile,"%f,%f,%f,%f,%d",&a,&b,&c,&d,&i) == 5) {
      stoppingpowerinfo->fe[0] = a;
      stoppingpowerinfo->fn[0] = b;
      stoppingpowerinfo->fe[1] = c;
      stoppingpowerinfo->fn[1] = d;
      stoppingpowerinfo->nchoice = i;
    } else {
      GetMessageDialog("Error reading stopping power file.\n");
      goto stopreading;
    }

    /* --- now comes the fun part, we need to read in the stopping powers
       --- but first we need to get memory for them --- */
    if (stoppingpowerinfo->stopen1 == NULL) {
      stoppingpowerinfo->stopen1 = (float *) 
	malloc(sizeof(float) * stoppingpowerinfo->numcoarsestop);
    } else {
      stoppingpowerinfo->stopen1 = (float *) 
	realloc(stoppingpowerinfo->stopen1,sizeof(float) * stoppingpowerinfo->numcoarsestop);
    }
    if (stoppingpowerinfo->stop1 == NULL) {
      stoppingpowerinfo->stop1 = (float *)
	malloc(sizeof(float) * stoppingpowerinfo->numcoarsestop);
    } else {
      stoppingpowerinfo->stop1 = (float *)
	realloc(stoppingpowerinfo->stop1,sizeof(float) * stoppingpowerinfo->numcoarsestop);
    }
    if (stoppingpowerinfo->stopen2 == NULL) {
      stoppingpowerinfo->stopen2 = (float *) 
	malloc(sizeof(float) * stoppingpowerinfo->numcoarsestop);
    } else {
      stoppingpowerinfo->stopen2 = (float *) 
	realloc(stoppingpowerinfo->stopen2,sizeof(float) * stoppingpowerinfo->numcoarsestop);
    }
    if (stoppingpowerinfo->stop2 == NULL) {
      stoppingpowerinfo->stop2 = (float *)
	malloc(sizeof(float) * stoppingpowerinfo->numcoarsestop);
    } else {
      stoppingpowerinfo->stop2 = (float *)
	realloc(stoppingpowerinfo->stop2,sizeof(float) * stoppingpowerinfo->numcoarsestop);
    }
    /* --- now that memory is allocated let's get the information --- */
    for (i = 0; i < stoppingpowerinfo->numcoarsestop; i++) {
      if (fscanf(infile,"%f,%f,%f,%f",&a,&b,&c,&d) == 4) {
	stoppingpowerinfo->stopen1[i] = a;
	stoppingpowerinfo->stop1[i] = b;
	stoppingpowerinfo->stopen2[i] = c;
	stoppingpowerinfo->stop2[i] = d;
      } else {
	GetMessageDialog("Error reading stopping powers.\n");
	goto stopreading;
      }
    }
    /* --- now we need to read in the number of residual nucleus production cross-sections --- */
    if (fscanf(infile,"%d",&i) == 1) {
      stoppingpowerinfo->numresidprodcross = i;
    } else {
      GetMessageDialog("Error reading the number of residual nucleus production crosssections.");
      goto stopreading;
    }
    /* --- must allocate memory for the cross section information --- */
    if (stoppingpowerinfo->crosssectionen == NULL) {
      stoppingpowerinfo->crosssectionen = (float *) 
	malloc(sizeof(float) * stoppingpowerinfo->numresidprodcross);
    } else {
      stoppingpowerinfo->crosssectionen = (float *) 
	realloc(stoppingpowerinfo->crosssectionen,
		sizeof(float) * stoppingpowerinfo->numresidprodcross);
    }
    if (stoppingpowerinfo->crosssection == NULL) {
      stoppingpowerinfo->crosssection = (float *) 
	malloc(sizeof(float) * stoppingpowerinfo->numresidprodcross);
    } else {
      stoppingpowerinfo->crosssection = (float *) 
	realloc(stoppingpowerinfo->crosssection,
		sizeof(float) * stoppingpowerinfo->numresidprodcross);
    }

    /* --- now we need to read in the cross-section tables and corresponding energies --- */
    i = 0;
    fgets(dummystr,160,infile);
    while (i < stoppingpowerinfo->numresidprodcross) {
      if (fgets(dummystr,160,infile) == NULL) {
	GetMessageDialog("Error reading residual cross section energies.\n");
	goto stopreading;
      }
      charpointer = (char *) &dummystr;
      while ((charpointer != NULL) && (i < stoppingpowerinfo->numresidprodcross)) {
	if (sscanf(charpointer,"%f",&a) == 1) {
	  stoppingpowerinfo->crosssectionen[i] = a;
	  i++;
	  //printf("%f\n",a);
	  charpointer = (char *) strstr(charpointer,",");
	  if (charpointer != NULL) 
	    charpointer = (char *) &charpointer[1];
	} else {
	  GetMessageDialog("Error reading the residual cross section energies.\n");
	}
      }
    }    
    i = 0;
    while (i < stoppingpowerinfo->numresidprodcross) {
      if (fgets(dummystr,160,infile) == NULL) {
	GetMessageDialog("Error reading residual cross section energies.\n");
	goto stopreading;
      }
      charpointer = (char *) &dummystr;
      while ((charpointer != NULL) && (i < stoppingpowerinfo->numresidprodcross)) {
	if (sscanf(charpointer,"%f",&a) == 1) {
	  stoppingpowerinfo->crosssection[i] = a;
	  //printf("%f\n",a);
	  i++;
	  charpointer = (char *) strstr(charpointer,",");
	  if (charpointer != NULL) 
	    charpointer = (char *) &charpointer[1];
	} else {
	  GetMessageDialog("Error reading the residual cross section energies.\n");
	}
      }
    }
    stoppingpowerinfo->tardensity = stoppingpowerinfo->tarelthick / (float)
      stoppingpowerinfo->numtarel;

  stopreading:
    fclose(infile);
  }
}

/* MakeFeedingEntry
 *
 * Produces a frame with three entries in it for population, tauentry, and tsfentry
 * according to the level and position fed to it.
 */
GtkWidget *MakeFeedingEntry(int level, int pos)
{
  GtkWidget *localframe;
  GtkWidget *localhbox;
  GtkWidget *locallabel;
  char dummystr[40];
  
  sprintf(dummystr,"%d,%d",level+1,pos+1);
  localframe = gtk_frame_new(dummystr);

  localhbox = gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(localframe),localhbox);

  locallabel = gtk_label_new("Pop = ");
  gtk_container_add(GTK_CONTAINER(localhbox),locallabel);
  population[level][pos] = gtk_entry_new();
  gtk_widget_set_usize(population[level][pos],25,22);
  gtk_container_add(GTK_CONTAINER(localhbox),population[level][pos]);
  locallabel = gtk_label_new("  tauentry = ");
  gtk_container_add(GTK_CONTAINER(localhbox),locallabel);
  tauentry[level][pos] = gtk_entry_new();
  gtk_widget_set_usize(tauentry[level][pos],25,22);
  gtk_container_add(GTK_CONTAINER(localhbox),tauentry[level][pos]);
  locallabel = gtk_label_new("ps  tauentry sf = ");
  gtk_container_add(GTK_CONTAINER(localhbox),locallabel);
  tsfentry[level][pos] = gtk_entry_new();
  gtk_widget_set_usize(tsfentry[level][pos],25,22);
  gtk_container_add(GTK_CONTAINER(localhbox),tsfentry[level][pos]);
  locallabel = gtk_label_new("ps");
  gtk_container_add(GTK_CONTAINER(localhbox),locallabel);

  return(localframe);
}

/* OpenFeedingWindow
 *
 * Creates a window to enter feeding information into
 */
void OpenFeedingWindow()
{
  GtkWidget *localwindow;
  GtkWidget *localtable;
  GtkWidget *localframe;
  GtkWidget *localbutton;
  int i,j;

  for (i = 0; i < 10; i++) {
    for (j = 0; j < 4; j++) {
      population[i][j] = NULL;
      tauentry[i][j] = NULL;
      tsfentry[i][j] = NULL;
    }
  }

  localwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(localwindow),"Feeding Information");
  
  localtable = gtk_table_new(5,10,FALSE);
  gtk_container_add(GTK_CONTAINER(localwindow),localtable);
  gtk_table_set_row_spacing(GTK_TABLE(localtable),0,2);
  gtk_table_set_col_spacing(GTK_TABLE(localtable),2,0);
  localframe = MakeFeedingEntry(0,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,2,3,9,10,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(1,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,0,2,8,9,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(1,1);
  gtk_table_attach(GTK_TABLE(localtable),localframe,3,5,8,9,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(2,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,0,1,7,8,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(2,1);
  gtk_table_attach(GTK_TABLE(localtable),localframe,1,2,7,8,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(2,2);
  gtk_table_attach(GTK_TABLE(localtable),localframe,3,4,7,8,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(2,3);
  gtk_table_attach(GTK_TABLE(localtable),localframe,4,5,7,8,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(3,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,0,1,6,7,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(3,1);
  gtk_table_attach(GTK_TABLE(localtable),localframe,1,2,6,7,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(3,2);
  gtk_table_attach(GTK_TABLE(localtable),localframe,3,4,6,7,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(3,3);
  gtk_table_attach(GTK_TABLE(localtable),localframe,4,5,6,7,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(4,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,0,1,5,6,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(4,1);
  gtk_table_attach(GTK_TABLE(localtable),localframe,1,2,5,6,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(4,2);
  gtk_table_attach(GTK_TABLE(localtable),localframe,3,4,5,6,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(4,3);
  gtk_table_attach(GTK_TABLE(localtable),localframe,4,5,5,6,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(5,0);
  gtk_table_attach(GTK_TABLE(localtable),localframe,0,1,4,5,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(5,1);
  gtk_table_attach(GTK_TABLE(localtable),localframe,1,2,4,5,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(5,2);
  gtk_table_attach(GTK_TABLE(localtable),localframe,3,4,4,5,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localframe = MakeFeedingEntry(5,3);
  gtk_table_attach(GTK_TABLE(localtable),localframe,4,5,4,5,
		   GTK_EXPAND | GTK_SHRINK,
		   GTK_EXPAND | GTK_SHRINK,
		   0,0);
  localbutton = gtk_button_new_with_label("Read Feeding File");
  gtk_table_attach(GTK_TABLE(localtable),localbutton,0,1,0,1,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   0,0);
  gtk_signal_connect(GTK_OBJECT(localbutton),"clicked",
		     GTK_SIGNAL_FUNC(FeedingFileReadCallback),
		     NULL);
  localbutton = gtk_button_new_with_label("Write Feeding File");
  gtk_table_attach(GTK_TABLE(localtable),localbutton,1,2,0,1,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   0,0);
  gtk_signal_connect(GTK_OBJECT(localbutton),"clicked",
		     GTK_SIGNAL_FUNC(FeedingFileSaveCallback),
		     NULL);
  localbutton = gtk_button_new_with_label("Clear");
  gtk_table_attach(GTK_TABLE(localtable),localbutton,2,3,0,1,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   0,0);
  gtk_signal_connect(GTK_OBJECT(localbutton),"clicked",
		     GTK_SIGNAL_FUNC(FeedingClear),
		     NULL);
  
  
  gtk_widget_show_all(localwindow);
}

/* FeedingFileSaveCallback
 *
 * Callback function which saves the current feeding information
 */
void FeedingFileSaveCallback(GtkWidget *widget, gpointer *data)
{
  GetFilename("Save Feeding Information",FeedingFileSave);
}


/* FeedingFileSave
 *
 * Saves the current feeding information
 */
void FeedingFileSave(char *sFilename)
{
  FILE *outfile;
  char dummystr[80];
  int i,j,k,l,m;
  float a,b,c,d,e;
  float popout,sfout,tauentryout;

  if ((outfile = fopen(sFilename,"w")) != NULL) {
    for (i = 0; i < 10; i++) {
      for (j = 0; j < 4; j++) {
	if ((population[i][j] != NULL) || (tauentry[i][j] != NULL) || (tsfentry[i][j] != NULL)) {
	  if (sscanf(gtk_entry_get_text(GTK_ENTRY(population[i][j])),"%f",&a) == 1) {
	    popout = a;
	  } else {
	    popout = 0;
	  }
	  if (sscanf(gtk_entry_get_text(GTK_ENTRY(tauentry[i][j])),"%f",&a) == 1) {
	    tauentryout = a;
	  } else {
	    tauentryout = 0;
	  }
	  if (sscanf(gtk_entry_get_text(GTK_ENTRY(tsfentry[i][j])),"%f",&a) == 1) {
	    sfout = a;
	  } else {
	    sfout = 0;
	  }
	  if ((popout != 0) || (tauentryout != 0) || (sfout != 0))
	    fprintf(outfile,"%d,%d,%f,%f,%f\n",i,j,popout,sfout,tauentryout);
	}
      }
    }

    fclose(outfile);
  }
}

/* FeedingFileReadCallback
 *
 * Callback function which saves the current feeding information
 */
void FeedingFileReadCallback(GtkWidget *widget, gpointer *data)
{
  GetFilename("Read Feeding Information",FeedingFileRead);
}


/* FeedingFileRead
 *
 * Saves the current feeding information
 */
void FeedingFileRead(char *sFilename)
{
  FILE *infile;
  char dummystr[80];
  int i,j,k,l,m;
  float a,b,c,d,e;
  float popout,sfout,tauentryout;

  if ((infile = fopen(sFilename,"r")) != NULL) {

    while(fgets(dummystr,80,infile) != NULL) {
      if(sscanf(dummystr,"%d,%d,%f,%f,%f",&i,&j,&a,&b,&c) == 5) {
	if (population[i][j] != NULL) {
	  sprintf(dummystr,"%f",a);
	  gtk_entry_set_text(GTK_ENTRY(population[i][j]),dummystr);
	}
	if (tsfentry[i][j] != NULL) {
	  sprintf(dummystr,"%f",b);
	  gtk_entry_set_text(GTK_ENTRY(tsfentry[i][j]),dummystr);
	}
	if (tauentry[i][j] != NULL) {
	  sprintf(dummystr,"%f",c);
	  gtk_entry_set_text(GTK_ENTRY(tauentry[i][j]),dummystr);
	}

      }
    }

    fclose(infile);
  }
}

/* FeedingClear
 *
 * Callback function which saves the current feeding information
 */
void FeedingClear(GtkWidget *widget, gpointer *data)
{
  int i,j,k;

  for (i = 0; i < 10; i++) {
    for (j = 0; j < 4; j++) {
      if (population[i][j] != NULL)
	gtk_entry_set_text(GTK_ENTRY(population[i][j]),"");
      if (tauentry[i][j] != NULL)
	gtk_entry_set_text(GTK_ENTRY(tauentry[i][j]),"");
      if (tsfentry[i][j] != NULL)
	gtk_entry_set_text(GTK_ENTRY(tsfentry[i][j]),"");
    }
  }
}

/* ReadFitsSetup
 *
 * Reads the last little bit of information needed to do a fits calculation
 */
void ReadFitsSetup(char *sFilename)
{
  int i,j,k;
  FILE *infile;
  float a,b,c;
    
  if ((infile = fopen(sFilename,"r")) != NULL) {
    
    if (stoppingpowerinfo == NULL) {
      stoppingpowerinfo = (struct stoppingpower *) malloc(sizeof(struct stoppingpower));
      stoppingpowerinfo->rectheta = NULL;
      stoppingpowerinfo->recphi = NULL;
      stoppingpowerinfo->recprobtheta = NULL;
      stoppingpowerinfo->projstop = NULL;
      stoppingpowerinfo->stopen1 = NULL;
      stoppingpowerinfo->stop1 = NULL;
      stoppingpowerinfo->stopen2 = NULL;
      stoppingpowerinfo->stop2 = NULL;
      stoppingpowerinfo->crosssectionen = NULL;
      stoppingpowerinfo->crosssection = NULL;
    }

    if (fscanf(infile,"%f,%f,%f",&a,&b,&c) == 3) {
      stoppingpowerinfo->theta = a;
      stoppingpowerinfo->a2 = b;
      stoppingpowerinfo->a4 = c;
    } else {
      GetMessageDialog("Error reading detector angle information.\n");
      goto stopreading;
    }
    if (fscanf(infile,"%f,%f,%f,%d,%d",&a,&b,&c,&i,&j) == 5) {
      stoppingpowerinfo->r = a;
      stoppingpowerinfo->h = b;
      stoppingpowerinfo->d = c;
      stoppingpowerinfo->nrad = i;
      stoppingpowerinfo->nang = j;
    } else {
      GetMessageDialog("Error reading detector geometry information.");
      goto stopreading;
    }

  stopreading:
    fclose(infile);
  }

}

/* Fit
 *
 * Does the fit according to the old system.  For more details see fits.f
 * in ~tabor/dsam/.
 */
void Fit()
{
  /* --- local variables --- */

  int i,j,k,l,m,n;
  float a,b,c,d,e,f;
  float imax,jmax;
  int igam;
  int nchns;
  float sumg;
  float reso;
  float gsq;
  float sm,gn,g;
  float z123,zt23,zb23;
  float sqt,sqb,eavg,aa2,z2;
  float thetd,sumslt,costh;
  float slt,angt,eltime;
  float iflag;
  float ebeam,sigmax,ebi,sigmr,er,rnum;
  float ebf;
  int ij,il,ih,igs,jab;
  int jj;
  float dtime,dtime1,t,dtime2,t1,t2,er1,er2,error;
  char dummystr[80];
  float sumth,anu,denom,argu;


  /* --- function --- */
  //printf("2 tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
  //	 fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);

  /* --- we have hopefully already read in the stoppng power file --- */
  /* --- guess we can check --- */
  g_return_if_fail(stoppingpowerinfo != NULL);
  /* --- we will also need to have calibrations set
     --- the UseCalibration function returns -1 if there are no
     --- calibrations --- */
  g_return_if_fail(UseCalibration(markers[0]) != -1);

  WriteMainText("\nStarting Fit.\n");
  //printf("Starting Fit.\n");

  /* --- now we need to get the experimental spectrum from the
     --- histogram values between the markers (inclusive) --- */
  qsort(&markers,2,sizeof(int),CompareInts); // get the first to markers in the correct order
  fitsstuff->ichn = markers[1] - markers[0] + 1;
  for ( i = 0; i < fitsstuff->ichn; i++) {
    fitsstuff->ispece[i] = (float) *(histloc[spectra] + markers[0] + i);
  }

  //printf("3 tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
  //	 fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);

  //printf("Got Histogram data.\n");

  /* --- if there is information for efficiency correction
     --- we can use it --- */
  if (UseEfficiency(fitsstuff->ispece[0],UseCalibration(markers[0])) != -1) {
    for (i = 0; i < fitsstuff->ichn; i++) {
      fitsstuff->ispece[i] = UseEfficiency(fitsstuff->ispece[i],UseCalibration(markers[0] + i));
    }
  }

  //printf("4 tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
  //	 fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);
  
  //printf("If avalible corrected histogram data for efficiency.\n");
  /* --- there now we have the correct histogram information --- */

  imax = jmax = 0;

  /* --- now we need to get the population parameters --- */
  for ( i = 0; i < 4; i ++) {
    for ( j = 0; j < 10; j ++) {
      if ((population[i][j] != NULL) || (tauentry[i][j] != NULL) || (tsfentry[i][j] != NULL)) {
	if (sscanf(gtk_entry_get_text(GTK_ENTRY(population[i][j])),"%f",&a) == 1) {
	  fitsstuff->pop[i][j] = a;
	} else {
	  fitsstuff->pop[i][j] = 0;
	}
	if (sscanf(gtk_entry_get_text(GTK_ENTRY(tauentry[i][j])),"%f",&a) == 1) {
	  fitsstuff->tau[i][j] = a;
	} else {
	  fitsstuff->tau[i][j] = 0;
	}
	if (sscanf(gtk_entry_get_text(GTK_ENTRY(tsfentry[i][j])),"%f",&a) == 1) {
	  fitsstuff->tsf[i][j] = a;
	} else {
	  fitsstuff->tsf[i][j] = 0;
	}
	if ((fitsstuff->pop[i][j] != 0) || (fitsstuff->tau[i][j] != 0) || (fitsstuff->tsf[i][j] != 0)) {
	  imax = i;
	  jmax = j;
	}
      } else {
	fitsstuff->pop[i][j] = 0;
	fitsstuff->tau[i][j] = 0;
	fitsstuff->tsf[i][j] = 0;
      }
      fitsstuff->tsf[i][j] = fitsstuff->tsf[i][j] * pow(10,-12);
      fitsstuff->tau[i][j] = fitsstuff->tau[i][j] * pow(10,-12);
    } // end j loop
  } // end i loop


  //printf("4 tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
  //	 fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);

  for (i = 0; i < 30; i++)
    fitsstuff->taul[i] = 0;

  if (fitsstuff->pop[0][0] < 1) fitsstuff->pop[0][0] = 1;

  /* --- now we have the population parameters --- */

  //printf("Got Population Parameters.\n");

  fitsstuff->igmax = 1;
  igam = 0;

  //printf("5 tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
  // fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);

  fitsstuff->nlt = (int) 
    ((float) (fitsstuff->tauup - fitsstuff->taulo) / (float) (fitsstuff->dtau) + 1);
  if (fitsstuff->nlt > 29) fitsstuff->nlt = 29;
  fitsstuff->nlts = fitsstuff->nlt + 1;
  for ( i = 0; i < fitsstuff->nlts; i++) {
    j = i - 1;
    fitsstuff->taul[i] = (fitsstuff->taulo + j * fitsstuff->dtau) * pow(10,-12);
  }

  /* --- now we need to get the stop peak info and egam --- */
  // printf("Getting stop peak possition and egam.\n");
  
  qsort(&markers[2],2,sizeof(int),CompareInts);
  /* --- later the egam variable should be modified to
     --- vary with stop peak possition --- */
  fitsstuff->egam = UseCalibration(markers[2]);
  
  fitsstuff->pklo = markers[2] - markers[0];
  fitsstuff->pkhi = markers[3] - markers[0];
  fitsstuff->dpeak = (float) (markers[3] - markers[2]) / (float) 9;
  fitsstuff->np1 = (int) 
    (float) (fitsstuff->pkhi - fitsstuff->pkhi) / (float) (fitsstuff->dpeak + 1);
  if (fitsstuff->np1 > 10) fitsstuff->np1 = 10;
  fitsstuff->pkhi = (fitsstuff->np1 - 1) * fitsstuff->dpeak + fitsstuff->pklo;

  /* --- get the background and the calibration from the globals --- */
  //printf("Getting background.\n");

  fitsstuff->ibkg = background;
  if(calibration[spectra][1] != 0) {
    fitsstuff->ecal = calibration[spectra][1];
  } else {
    fitsstuff->ecal = globalcalibration[1];
  }
  
  /* --- while stoppingpowerinfo->nrad and nang are already in memory 
     --- we need to keep them in appropriate ranges --- */
  if (stoppingpowerinfo->nrad > 10) stoppingpowerinfo->nrad = 10;
  if (stoppingpowerinfo->nang > 10) stoppingpowerinfo->nang = 10;

  /* --- now let's actually do the lineshape fitting --- */
  //printf("Starting to fit the lineshape.\n");

  sumg = 0;
  reso = stoppingpowerinfo->detres[0] +
    stoppingpowerinfo->detres[1] * fitsstuff->egam + 
    stoppingpowerinfo->detres[2] * pow(fitsstuff->egam,2);
  g_return_if_fail(reso != 0);
  
  for ( i = 0; i < 160; i++) {
    gsq = pow(((float) i / ((reso * 6.006) / fitsstuff->ecal)) , 2);
    sumg = sumg + exp(-gsq);
  }
  
  sm = 1 + 2 * sumg;
  gn = (float) 1 / (float) sm;
  for ( i = 0; i < 160; i++) {
    j = i - 1;
    g = (float) j/ ((6.006 * reso) / fitsstuff->ecal);
    gsq = pow(g,2);
    fitsstuff->gauss[i] = gn * exp(-gsq);
  }
  
  z123 = pow(stoppingpowerinfo->recz, 0.6666667);
  zt23 = pow(stoppingpowerinfo->targz, 0.6666667);
  zb23 = pow(stoppingpowerinfo->backz, 0.6666667);
  sqt = sqrt(z123 + zt23);
  sqb = sqrt(z123 + zb23);
  fitsstuff->epfac[0] = (32536 * stoppingpowerinfo->targa) / 
    (stoppingpowerinfo->recz * stoppingpowerinfo->targz * 
     (stoppingpowerinfo->targa + stoppingpowerinfo->reca) * sqt);
  fitsstuff->facnuke[0] =
    (5.1048 * stoppingpowerinfo->recz * stoppingpowerinfo->targz * stoppingpowerinfo->reca) /
    (stoppingpowerinfo->targa * sqt *
     (stoppingpowerinfo->targa * stoppingpowerinfo->reca));
  if (stoppingpowerinfo->tarstopnorm[1] == 0) {
    fitsstuff->epfac[1] = 0;
    fitsstuff->facnuke[1] = 0;
  } else {
    fitsstuff->epfac[1] = (32536 * stoppingpowerinfo->backa) / 
      (stoppingpowerinfo->recz * stoppingpowerinfo->backz * sqb *
       (stoppingpowerinfo->backa + stoppingpowerinfo->reca));
    fitsstuff->facnuke[1] = 
      (5.1048 * stoppingpowerinfo->recz * stoppingpowerinfo->backz * stoppingpowerinfo->reca) / 
      (stoppingpowerinfo->backa * sqb * (stoppingpowerinfo->backa + stoppingpowerinfo->reca));
  }

  if (stoppingpowerinfo->blauswitch) {
    eavg = (float) (stoppingpowerinfo->projmass * stoppingpowerinfo->beamenergy) /
      (float) (stoppingpowerinfo->projmass * stoppingpowerinfo->targa);
    if (stoppingpowerinfo->tarstopnorm[1] == 0) {
      aa2 = stoppingpowerinfo->targa;
      z2 = stoppingpowerinfo->targz;
    } else {
      aa2 = stoppingpowerinfo->backa;
      z2 = stoppingpowerinfo->backz;
    }
    
    /* --- this is where cobar is called
       --- I'm calling it FitsAvCos --- */
    //printf("about to call FitsAvCos.\n");
    FitsAvCos(stoppingpowerinfo->reca,stoppingpowerinfo->recz,aa2,z2,eavg);
  }

  if ((stoppingpowerinfo->nrad == 1) && (stoppingpowerinfo->numrectheta == 1)) {
    fitsstuff->probn[0][0][0][0] = 1;
    stoppingpowerinfo->numrecphi = 1;
    stoppingpowerinfo->nang = 1;
    thetd = stoppingpowerinfo->theta * 3.1415926 / (float) 180;
    fitsstuff->cst[0][0][0][0] = cos(thetd);
  } else {
    /* --- this is where detgeom is called
       --- I'm going to call it DetectorGeomretry --- */
    //printf("About to call Detector Geometry.\n");
    DetectorGeometry();
    //printf("Done with Detector Geometry.\n");

    sumslt = 0;
    costh = cos(stoppingpowerinfo->theta / 57.296);
      for (i = 0; i < stoppingpowerinfo->nang; i++) {
	for ( j = 0; j < stoppingpowerinfo->nrad; j++) {
	  for ( k = 0; k < stoppingpowerinfo->numrecphi; k++) {
	    for (l = 0; l < stoppingpowerinfo->numrectheta; l++) {
	      sumslt = sumslt + fitsstuff->probn[i][j][l][k] *
	      pow((fitsstuff->cst[i][j][l][k] - costh),2);
	    }
	  }
	}
      }
    slt = sqrt(sumslt);
    angt = 57.296 * acos(slt);
  }
  fitsstuff->acn = stoppingpowerinfo->projmass + stoppingpowerinfo->targa;
  
  /* --- now we have egam and stop peak information --- */
  
  /* --- guess we'll try to calculate the spectrum --- */

  eltime = 0;
  for ( i = 0; i < 30; i++) {
    for ( j = 0; j < 1000; j++) {
      fitsstuff->specn[i][j] = 0;
    }
  }

  iflag = 0;
  ebeam = stoppingpowerinfo->beamenergy;

  /* --- this is where sigma is called --- */
  /* --- I call it CrossSection --- */
  //printf("About to call CrossSection.\n");
  sigmax = CrossSection(ebeam);
  //printf("Called CrossSection.\n");
  for ( i = 0; i < stoppingpowerinfo->numtarel; i++) {
    sprintf(dummystr,"on %d of %d target elements.\n",i,(int) stoppingpowerinfo->numtarel);
    WriteMainText(dummystr);
    ebi = ebeam;
    //printf("About to call cross section for the %d time.\n",i+2);
    sigmr = CrossSection(ebi);
    //printf("Executed CrossSection for the %d time.\n",i+2);
    if (sigmr < 0) goto crossdone;
    /* --- this is where kinemat is called --- */
    //printf("About to call Kinemat.\n");
    Kinemat(ebi,er);
    //printf("Called Kinemat.\n");
    for ( j = 0; j < 7; j++) {
      fitsstuff->rnumi[j] = fitsstuff->eprob[j] * rnum * sigmr/sigmax;
    }
    jj = i;
    /* --- this is where shift is called --- */
    //printf("About to call Shift.\n");
    Shift(er,jj);
    //printf("Called Shift.\n");
    iflag = 0;
    /* --- this is where stopp is called
       --- I call it ProjectileStop --- */
    //printf("About to call ProjectileStop.\n");
    ebeam = ProjectileStop(ebi);
    //printf("Called ProjectileStop.\n");
    ebf = ebeam;
    for ( j = 0; j < fitsstuff->nlt; j++) {
      for ( k = 0; k < 1000; k++) {
	fitsstuff->specn[j][k] = fitsstuff->specn[j][k] + fitsstuff->spec[j][k] / rnum;
      } // done k loop
    } // done j loop
    if (ebf < stoppingpowerinfo->minbeamen) goto crossdone;
  } // done i loop
 crossdone:
  /* --- now we have two spectr of counts vs. energy shift
   * in 1 keV steps, with specn[ig,500] = unshifted peaks.
   * we need to fold in detector resolution --- */

  for ( i = 0; i < fitsstuff->nlt; i++) {
    for ( j = 0; j < 1200; j++) {
      fitsstuff->spect[i][j] = 0;
    }
  }
  
  for ( i = 0; i < fitsstuff->nlt; i++) {
    for ( j = 0; j < 1000; j++) {
      ij = i + 160;
      if (fitsstuff->specn[i][j] > 0) {
	il = ij - 160;
	ih = ij + 160;
	for ( k = il; k <= ih; k++) {
	  igs = k - ij;
	  jab = abs(igs) + 1;
	  jj = j - 60;
	  if ((jj > 0) || (jj < 1200)) {
	    fitsstuff->spect[i][jj] = 
	      fitsstuff->spect[i][jj] * fitsstuff->specn[i][jj] * fitsstuff->gauss[jab];
	  }
	} 
      } else { 
	goto nexti;
      }
    }
  nexti:
    sumth = 0;
    for ( j = 0; j < 1200; j++) {
      sumth = sumth + fitsstuff->spect[i][j];
    }
  }

  /* --- now we have to call ChiSqr --- */
  printf("About to call Chisqr.\n");
  ChiSqr();
  printf("Called Chisqr.\n");

  if (fitsstuff->bestchi[0] != 0) {
    for (i = 0; i < 3; i++) {
      fitsstuff->qft[i] = fitsstuff->bestchi[i] / fitsstuff->bestchi[0];
    }
    anu = fitsstuff->qft[2] - fitsstuff->qft[1];
    denom = fitsstuff->qft[2] - 2 * fitsstuff->qft[1] + fitsstuff->qft[0];
    if (denom != 0) {
      nchns = fitsstuff->ichn - fitsstuff->ich1 + 1;
      argu = fabs((float)2 / (float) (nchns * denom));
      if (fitsstuff->igmax == 1) {
	dtime = fitsstuff->bestt1[2] - fitsstuff->bestt1[1];
	if (dtime == 0) dtime = fitsstuff->dtau / (1E-12);
	t = fitsstuff->bestt1[2] - dtime * (anu / (denom + 0.5));
	error = dtime * sqrt(argu);
	t = t / (1E-12);
	error = error / (1E-12);
      } else {
	dtime1 = fitsstuff->bestt1[2] - fitsstuff->bestt1[1];
	dtime2 = fitsstuff->bestt2[2] - fitsstuff->bestt2[1];
	t1 = fitsstuff->bestt1[2] - dtime1 * (anu / (denom + 0.5));
	t2 = fitsstuff->bestt2[2] - dtime2 * (anu / (denom + 0.5));
	if (dtime1 == 0) dtime1 = (fitsstuff->dtau * 1E-12);
	if (dtime2 == 0) dtime2 = (fitsstuff->dtau * 1E-12);
	er1 = dtime1 * sqrt(argu);
	er2 - dtime2 * sqrt(argu);
	t1 = t1 / (1E-12);
	t2 = t2 / (1E-12);
	er1 = er1 / (1E-12);
	er2 = er2 / (1E-12);
      }
    }
  }
  for (i = 0; i < 10; i++) {
    sprintf(dummystr," %d: Back: %d  Peak: %.1f  Tau: %.2f  Chisqr: %.2f\n",
	    i+1,fitsstuff->bestbk[i],fitsstuff->peakpos[i],
	    fitsstuff->bestt1[i] / (1E-12),fitsstuff->bestchi[i]);
    WriteMainText(dummystr);
  }
  /* --- free appropriate memory --- */

  free(fitsstuff);
  fitsstuff = NULL;

}

/* FitsTauEntry
 *
 * Interprets the tau entry from the dialog
 */
void FitsTauEntry(GtkWidget *widget, GtkWidget *entry)
{
  float a,b,c;
  int test;
  char dummystr[80];

  /* --- make sure that we get the right input --- */
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(entry)),"%f %f %f",&a,&b,&c) == 3) {

  /* --- allocate memory --- */

    fitsstuff = (struct fitsvariables *) malloc(sizeof(struct fitsvariables));
    fitsstuff->taulo = a;
    fitsstuff->tauup = b;
    fitsstuff->dtau = c;
  printf("tauup = %.2f, taulo = %.2f, dtau = %.2f\n",
	 fitsstuff->tauup,fitsstuff->taulo,fitsstuff->dtau);
    Fit();
  } else {
    sprintf(dummystr,"Error reading tau parameters from entry.");
    GetMessageDialog(dummystr);
  }
}

/* FitsTauPrompt
 *
 * Prompts the user for the parameters for tau
 */
void FitsTauPrompt(GtkWidget *text)
{
  char dummystr[80];
  sprintf(dummystr,"Please enter Taulo Tauup, dTau.");
  gtk_text_insert(GTK_TEXT(text),NULL,NULL,NULL,dummystr,strnlen(&dummystr,80));
}

/* FitsAvCos
 *
 * Calculates the average cosine of multiple
 * scattering angle, cbar, due to nuclear scattering
 * of the recoils using Blaugrund's approx.
 */
void FitsAvCos(float a1, float z1, float a2, float z2, float er)
{
  float emax,arec;
  int i,j,k,l,m;
  float lphase;
  float apa,z123,z223;
  float r,g,am,d,ak;
  float dele,efac,delp,sumi,e,je,elp,dee,sqrel,den;

  emax = er;
  arec = a2;
  lphase = 1;
  if (emax > 5) emax = 5;
  apa = a1+a2;
  z123 = pow(z1,0.66667);
  z223 = pow(z2,0.66667);
  r = a1/a2;
  if ( r < 1) {
    g = 1 + 0.6666667 * r - 0.46667 * pow(r,2);
  } else {
    g = 0.6666667 + 8 / (15 * r);
  }
  am = 1630 * a1 * a2 / (z1 * z2 * sqrt(z123 + z223) * apa);
  d = pow((z123 + z223),0.75) * sqrt(pow (a1,3) * a2);
  ak = pow(z1,0.166667)*0.0793*sqrt(z1*z2*pow(apa,3)) / d;
  dele = 0.01;
  efac = 10.2 * am /(0.511 * a1);
  delp = dele * efac;
  sumi = 0;
  e = emax;
  je = 100 * e +0.5;
  while (je >= 0) {
    elp = e * efac;
    dee = ak *sqrt(elp);
    sqrel = - sqrt(elp);
    den = 0.6111 * exp(sqrel/1.919) * (1 - exp(sqrel/0.406));
    sumi = sumi + delp * den / ((den + dee) * elp);
    fitsstuff->cbar[(int)je] = exp(-0.5*g*sumi/r);
    e = e - dele;
    je = 100 * e + 0.5;
  }

}

/* DetectorGeomretry
 *
 * This subroutine produces an array of gamma ray
 * detection probablilities, PROBN(ig), for each
 * detector element, and an array of gamma ray 
 * angles through each detector element CST.  PROBN
 * and CST are also indexed by theta and phi of the
 * recoil direction.
 */
void DetectorGeometry()
{
  float thetad;
  float ssd,csd,rho1,delphi,delrho;
  float delphir,sump,rho,rhsq,dsq,resq,ael,domega,xel;
  float rd,xd,shot,arg,eff,phi,csi,ssi,xp,yp,zp;
  float xp2,yp2,zp2,densq,den,thetr;
  float ssr,csr,phir,xr,yr,zr,csp;
  float p2,p4,wt;
  int i,j,k,l;
  float thetd,cst,relsq;

  thetd = stoppingpowerinfo->theta * 3.1415926 / 180;
  ssd = sin(thetd);
  cst = cos(thetd);
  rho1 = stoppingpowerinfo->r * stoppingpowerinfo->d / 
    (stoppingpowerinfo->r + stoppingpowerinfo->d);
  delphi = 3.1415926 / (float) stoppingpowerinfo->nang;
  delrho = stoppingpowerinfo->r / stoppingpowerinfo->nrad;
  delphir = 6.2831853;  // 2 * pi
  
  sump = 0;
  for (i = 0; i < stoppingpowerinfo->nrad; i++) {
    rho = ((float)i - 0.5) * delrho;
    rhsq = pow(rho,2);
    dsq = pow(stoppingpowerinfo->d,2);
    relsq = dsq + rhsq;
    ael = 0.5 * (2 * rho + delrho) * delrho * delphi;
    domega = ael / relsq;
    if (rho < rho1) {
      xel = stoppingpowerinfo->h / stoppingpowerinfo->d * sqrt(relsq);
    } else {
      rd = pow( (stoppingpowerinfo->r - rho),2);
      xd = 1 + dsq / rhsq;
      shot = rd * xd;
      xel = sqrt(shot);
    }
    arg = 1000 * xel / fitsstuff->egam;
    eff = 1 - exp(-arg);
    for ( j = 0; j < stoppingpowerinfo->nang; j++) {
      phi = ((float)j - 0.5) * delphi;
      csi = cos(phi);
      ssi = sin(phi);
      xp = rho * csi * csd + stoppingpowerinfo->d * ssd;
      yp = rho * ssi;
      zp = stoppingpowerinfo->d * csd - rho * csi * ssd;
      xp2 = pow(xp,2);
      yp2 = pow(yp,2);
      zp2 = pow(zp,2);
      densq = xp2 + yp2 + zp2;
      den = sqrt(densq);
      xp = xp / den;
      yp = yp / den;
      zp = zp / den;
      /* --- xp,yp,zp is a unit vector pointing from the target to the current
	 --- detector segment --- */
      for (k = 0; k < stoppingpowerinfo->numrectheta; k++) {
	thetr = stoppingpowerinfo->rectheta[k];
	ssr = sin(thetr);
	csr = cos(thetr);
	for ( l = 0; l < stoppingpowerinfo->numrecphi; l++) {

	  phir = ((float) l - 0.5) * delphir;
	  xr = ssr * sin(phir);
	  yr = ssr * cos(phir);
	  zr = csr;

	  /* --- xr,yr,zr is a unit vector in the direction or the recoil
	     --- velocity.  The cos of the angle between the detector
	     --- and the recoil directions is given by the
	     --- dot product. --- */
	  csp = xr * xp + yr * yp + zr * zp;
	  p2 = AP2(csp);
	  p4 = AP4(csp);
	  wt = 1 + stoppingpowerinfo->a2 * p2 + stoppingpowerinfo->a4 * p4;
	  fitsstuff->cst[j][i][k][l] = csp;
	  fitsstuff->probn[j][i][k][l] = domega * eff * wt * 
	    stoppingpowerinfo->recprobtheta[k];
	  sump = sump + fitsstuff->probn[j][i][k][l];
	} // done l loop

      } // done k loop

    } // done j loop

  } // done i loop


  for (i = 0; i < stoppingpowerinfo->nrad; i++) {  
    for ( j = 0; j < stoppingpowerinfo->nang; j++) {
      for (k = 0; k < stoppingpowerinfo->numrectheta; k++) {
	for ( l = 0; l < stoppingpowerinfo->numrecphi; l++) {
	  fitsstuff->probn[j][i][k][l] = fitsstuff->probn[j][i][k][l] / sump;
	}
      }
    }
  }
}

/* AP2
 * short function used in Detector Geometry
 */
float AP2(float x)
{
  return((float)0.5 * (3 * pow(x,2) - 1));
}

/* AP4
 * short function used in Detector Geometry
 */
float AP4(float x)
{
  return(0.125 * (35 * pow(x,4) - 30 * pow(x,2) + 3));
}

/* CrossSection
 *
 * This subroutine calculates the residual nucleus
 * production cross-section for the input
 * energy by searching for the nearest
 * energy entries in the input tables, and 
 * performing a linear interpolation to obtain
 * the value for crossection
 */
float CrossSection(float beamenergy)
{

  int i,j,k,l,m;
  float elo,eup,sigma;
  int jlo,jhi;
  float ehi;

  WriteMainText("Calculating cross-section.\n");

  for (i = 0; i < stoppingpowerinfo->numresidprodcross; i++) {
    jlo = i;
    jhi = i+1;
    elo = stoppingpowerinfo->crosssectionen[jlo];
    ehi = stoppingpowerinfo->crosssectionen[jhi];
    if (beamenergy == elo) {
      return(stoppingpowerinfo->crosssection[jlo]);
    }
    if (beamenergy == ehi) {
      return(stoppingpowerinfo->crosssection[jhi]);
    }
    if ((beamenergy > elo) && ( beamenergy < ehi)) {
      /* --- twint is called, I call it LinearInterPol --- */
      return(LinearInterPol(elo,eup,stoppingpowerinfo->crosssection[jlo],
			    stoppingpowerinfo->crosssection[jhi],beamenergy));
    }
  }
}

/* LinearInterPol
 *
 * Returns a float to the y value of a linear interpolation
 * between two points for a given x
 */
float LinearInterPol(float x1, float x2, float y1, float y2, float x)
{
  float det,anum,bnum,deno;
  float a,b,c;

  deno = x1 - x2;
  anum = y1 * x2 - y2 * x2;
  bnum = y1 - y2;
  a = anum / deno;
  b = bnum / deno;
  return(a + b * x);
}

/* Kinemat
 *
 * This subroutine calculates the energy of the
 * recoiling nuclei, ER, for an input projectile
 * energy
 */
void Kinemat(float beamenergy, float recoilenergy)
{
  float ecn,betacn,eavg,sig,arg,argsq,ter,sum,earg,eargsq,eterm;
  int i,j,k;
  float acn,del,term,pnorm,rnumir,ei;
  
  ecn = beamenergy * stoppingpowerinfo->projmass / fitsstuff->acn;
  betacn = sqrt((2*ecn) / (931.5 * acn));
  eavg = 0.5 * stoppingpowerinfo->reca * 931.5 * pow(betacn,2);
  sig = del * eavg/1.4142;
  for ( i = 0; i < 7; i++) {
    fitsstuff->er[i] = eavg + (4 - i) * sig;
  }
  sum = 0;
  for (i = 0; i < 3; i++) {
    arg = (fitsstuff->er[i] - eavg) / (1.4142 * sig);
    argsq = pow(arg,2);
    term = exp(-argsq);
    sum = sum + term;
  }
  pnorm = 1 / ( 1 + 2 * sum);
  for ( i = 0; i < 7; i++) {
    earg = (fitsstuff->er[i] - eavg) / ( 1.4142 * sig);
    eargsq = pow(earg,2);
    eterm = exp(-eargsq);
    fitsstuff->eprob[i] = pnorm * eterm;
  }
}

/* Shift
 *
 * Takes the recoils produced in the jth target element 
 * with energy enow, and propagates them through the 
 * remianing target elmenets.  It returns two shifted 
 * spectra, spec without detector resolution folded in 
 * and spec (ig,500) = the unshifted peaks.
 */
void Shift(float enw, int jj)
{
  int i,j,k;
  float a,b,c;
  float tt;
  int flag;
  int kstop;
  float enow[7];
  float rnumir,ei;
  float time,dlg;
  int ik;
  float ef,dedx;
  float beti,betf,betavg,vavg,ttime;
  float *dkay;
  float beta,delgam,elt;
  float eavg;
  int je;
  float cosb;
  int jang,jrad,jthet,jphi;
  int ielo,ieup;
  int imax,ir;
  float frac,counts;
  char dummystr[80];

  WriteMainText("Propogating recoiling nuclei through target.\n");

  dkay = (float *) malloc(sizeof(float) * fitsstuff->nlt);

  enow[0] = enw;

  tt = 0;
  for ( i = 0; i < 30; i++) {
    for (j = 0; j < 1000; j++) {
      fitsstuff->spec[i][j];
    }
  }
  
  //printf("Initialized Shift varialbes.\n");

  /* --- loop through target elements --- */
  for ( i = 0; i < 7; i++) {
    sprintf(dummystr,"i=%d",i);
    WriteMainText(dummystr);
    //printf("Looping through target element %d.\n",i+1);
    rnumir = fitsstuff->rnumi[i];
    flag = 0;
    kstop = 0;
    for ( j = 0; j < 20000; j++) {
      
      ei = enow[i];

      /* --- check to see if the recoils have stopped --- */
      if (ei <= 0.001) {
	flag = 2;
	/* --- this is where depop is used 
	   --- I call it DepopulateLevels --- */
	//printf("About to call DepopulateLevels.\n");
	DepopulateLevels(time,flag);
	//printf("Called DepopulateLevels.\n");
	//	printf("1 nlt = %d.\n",(int)fitsstuff->nlt);
	for ( k = 0; k < (int) fitsstuff->nlt; k++) {
	  fitsstuff->spec[k][500] = fitsstuff->spec[k][500] + fitsstuff->dki[k];
	} // done k loop
	//printf("Done with spec.\n");
      } else {
	if (j > stoppingpowerinfo->numtarel) { ik = 1;
	} else {  ik = 0; }
	/* --- this is where stopr is called --- */
	/* --- I call it GetStoppingpower --- */
	//printf("About to call GetStoppingPower.\n");
	GetStoppingPower(ei,ik,&ef,kstop,&dedx);
	//printf("Called GetStoppingPower.\n");
	kstop = 1;
	if (ef < 0) ef = 0;
	beti = sqrt((2 * ei) / (931.5 * stoppingpowerinfo->reca));
	betf = sqrt((2 * ef) / (931.5 * stoppingpowerinfo->reca));
	betavg = 0.5 * (beti + betf);
	vavg = 2.99792458E10 * betavg;
	time = stoppingpowerinfo->tardensity / vavg;
	ttime = time / 1E-12;
	tt = tt + time;
	/* --- call DepopulateLevels again --- */
	//printf("About to call DepopulateLevels.\n");
	DepopulateLevels(time,flag);
	//printf("Called DepopulateLevels.\n");
	//printf("2 nlt = %d.\n",(int)fitsstuff->nlt);
	for ( k = 0; k < (int)fitsstuff->nlt; k++) {
	  dkay[k] = fitsstuff->dki[k];
	}
	//printf("Made it through dkay.\n");
	flag = 1;
	eavg = 0.5 * 931.502 * stoppingpowerinfo->reca * pow(betavg,2);
	je = 100 * eavg + 0.5;
	if ( je < 0 ) je = 1;
	if ((je > 500) || (stoppingpowerinfo->blauswitch != 1)){
	  cosb = 1;
	} else {
	  //printf("je = %d\n",(int) je);
	  cosb = fitsstuff->cbar[(int)je];
	}
	beta = betavg * cosb;
	delgam = 10 * fitsstuff->egam * beta;
	
	//printf("About to enter big loop. %d %d\n",i,j);
	for (jang = 0; jang < stoppingpowerinfo->nang; jang++) {
	  for (jrad = 0; jrad < stoppingpowerinfo->nrad; jrad++) {
	    for (jthet = 0; jthet < stoppingpowerinfo->numrectheta; jthet++) {
	      for (jphi = 0; jphi < stoppingpowerinfo->numrecphi; jphi++) {
		dlg = delgam * fitsstuff->cst[jang][jrad][jthet][jphi] / fitsstuff->ecal;
		if (dlg > 0) {
		  ielo = (int) dlg + 500;
		  ieup = ielo + 1;
		  if (ielo > 999) ielo = 999;
		  frac  = dlg - (float)( ielo - 500 );
		  for (k = 0; k < fitsstuff->nlt; k++) {
		    counts = dkay[k] * fitsstuff->probn[jang][jrad][jthet][jphi];
		    fitsstuff->spec[k][ielo] = fitsstuff->spec[k][ielo] + (1.0 - frac) * counts;
		    fitsstuff->spec[k][ieup] = fitsstuff->spec[k][ieup] + frac * counts;
		  }
		} else {
		  ieup = (int) dlg + 500;
		  ielo = ieup - 1;
		  if (ieup <2) ieup = 2;
		  frac = dlg + ( 500 - ieup);
		  for (k = 0; k < fitsstuff->nlt ;k++) {
		    counts = dkay[k] * fitsstuff->probn[jang][jrad][jthet][jphi];
		    fitsstuff->spec[k][ieup] = fitsstuff->spec[k][ieup] + (1.0 - frac) * counts;
		    fitsstuff->spec[k][ielo] = fitsstuff->spec[k][ielo] + frac * counts;
		  }
		}
	      } // done jphi loop
	    } // done jthet loop
	  } // done jrad loop
	} // done jang loop
	//printf("Done with big loop.\n");
      }
      enow[i] = ef;
      if ((j % 1000) == 0) {
	WriteMainText(".");	    
	/* --- keep the window from dieing --- */
	while (gtk_events_pending ()) {
	  gtk_main_iteration ();
	}
      }
    } // done j loop

  } // done i loop
  elt = tt / 1E-12;
  imax = stoppingpowerinfo->numtarel;
  free(dkay);
}

/* DepopulateLevels
 *
 * This subroutine sets up initial populations
 * of each level, then depopulates them
 * using the radioactive decay law, 
 * dn = n * time/tau, for each level.  It then
 * puts the decays from each level into
 * the level below it, and finally calculates
 * the number of decays from each of the
 * levels of interest and returns the values 
 * in fitsstuff->dki
 */
void DepopulateLevels(float time, int flag)
{
  float dnn;
  int i,j,k;
  float a,b,c;
  float sfi[4][10];
  float dn[4][11];
  float anumi[4][10];
  float anum2[30];
  float sf,tau,dnc[4];
  float dn2[2],sfl[2];
  float dnl2[30];
  float totdk[30],dnt;

  //  WriteMainText("Depopulating Levels.\n");

  //printf("Flag = %d.\n",flag);

  dnn = 0;
  if ( flag == 0) {
    /* --- set up initial populations --- */

    for (i = 0; i < 10; i ++) {
      k = 9 - i;
      if (k > 1) {
	for (j = 0; j < 4; j++) {
	  sfi[j][k] = fitsstuff->pop[j][k] * fitsstuff->rnumi[0] / fitsstuff->pop[0][0];
	} // done with j loop
      } else {
	sfi[j][k] = fitsstuff->rnumi[0] * 
	  (fitsstuff->pop[j][k] - fitsstuff->pop[j][k+1]) / fitsstuff->pop[0][0];
      }
    } // done with i loop
    sfi[1][0] = (fitsstuff->pop[1][0] - fitsstuff->pop[2][0] - fitsstuff->pop[2][1]) *
      fitsstuff->rnumi[0] / fitsstuff->pop[0][0];
    sfi[1][1] = (fitsstuff->pop[1][1] - fitsstuff->pop[2][2] - fitsstuff->pop[2][3]) * 
      fitsstuff->rnumi[0] / fitsstuff->pop[0][0];
    sfi[0][0] = (fitsstuff->pop[0][0] - fitsstuff->pop[1][0] - fitsstuff->pop[1][1]) *
      fitsstuff->rnumi[0] / fitsstuff->pop[0][0];
    if (fitsstuff->pop[0][1] > 0) {
      sfi[0][0] = (fitsstuff->pop[0][0] - fitsstuff->pop[0][1]) * fitsstuff->rnumi[0] / fitsstuff->pop[0][0];
    }
    for ( i = 0; i < 4; i++) {
      dn[i][10] = 0;
      for ( j = 0; j < 10; j++) {
	dn[i][j] = 0;
	anumi[i][j] = 0;
      }
    }
    for ( i = 0; i < 30; i++) anum2[i] = 0;
    //printf("Got initial populations.\n");

  }  // done with flag == 0
  if (flag != 2) {
    /* --- now we have the initial populations, calculate the number
       --- of decays from each level, ignoring levels that have
       --- already decayed away. --- */
    sf = 0;
    for ( i = 0; i < 4; i++) {
      for ( k = 2; k < 10; k++) {
	j = 11 - k;
	if (sfi[i][j] < 0) {
	  sf = 0;
	} else {
	  if (fitsstuff->tsf[i][j] > 0) {
	    sf = sfi[i][j] * ( 1 - exp(-time / fitsstuff->tsf[i][j]));
	    sfi[i][j] = sfi[i][j] - sf;
	    if (sfi[i][j] < 0) sfi[i][j] = 0;
	  } else {
	    sf = sfi[i][j];
	    sfi[i][j] = 0;
	  }
	}
	anumi[i][j] = anumi[i][j] + sf + dn[i][j+1];
	if (fitsstuff->tau[i][j] > 0) {
	  dn[i][j] = anumi[i][j] * (1 - exp(-time/fitsstuff->tau[i][j]));
	  anumi[i][j] = anumi[i][j] - dn[i][j];
	  if (anumi[i][j] < 0) anumi[i][j] = 0;
	} else {
	  dn[i][j] = anumi[i][j];
	  anumi[i][j] = 0;
	}
      } // done with k loop
      dnc[i] = dn[2][i];
    }  // done with i loop
    
    /* --- finished decaying the highest cascade chains, dn[j] j = 1,4
       --- are the # of decays from each chain. --- */
    
    dnn = 0;
    sf = 0;
    for ( i = 0; i < 2; i++) {
      if (sfi[i][1] > 0) {
	sf = 0;
      } else {
	if(fitsstuff->tsf[i][1] > 0) {
	  sf = sfi[i][1] * (1 - exp(-time/ fitsstuff->tsf[i][1]));
	  sfi[i][1] = sfi[i][1] - sf;
	  if (sfi[i][1] < 0) sfi[i][1] = 0;
	} else {
	  sf = sfi[i][1];
	  sfi[i][1] = 0;
	}
      }
      if ( i == 0) {
	dnn = dnc[0] + dnc[1];
      } else {
	dnn = dnc[2] + dnc[3];
      }
      anumi[i][1] = anumi[i][1] + sf + dnn;
      if (fitsstuff->tau[i][1] > 0) {
	dn2[i] = anumi[i][1] * ( 1 - exp( - time / fitsstuff->tau[i][1]));
	anumi[i][1] = anumi[i][1] - dn2[i];
	if (anumi[i][1] < 0) anumi[i][1] = 0;
      } else {
	dn2[i] = anumi[i][1];
	anumi[i][1] = 0;
      }
    }
    
    /* --- now we have finished decaying the two levels feeding the highest lying
       --- level of interest.  dn2 are the number of decays from each feeding level. --- */
    
    if (fitsstuff->tsf[0][0] > 0) {
      sfl[0] = sfi[0][0] * (1 - exp( - time / fitsstuff->tsf[0][0]));
      sfi[0][0] = sfi[0][0] - sfl[0];
      if (sfi[0][0] < 0) sfi[0][0] = 0;
    } else {
      sfl[0] = sfi[0][0];
      sfi[0][0] = 0;
    }
    
    for (i = 0; i < fitsstuff->nlt; i++) {
      anum2[i] = anum2[i] + sfl[1] + dn2[0] + dn2[1];
      if (fitsstuff->taul[i] > 0) {
	dnl2[i] = anum2[i] * (1 - exp( - time / fitsstuff->taul[i]));
	anum2[i] = anum2[i] - dnl2[i];
      } else {
	dnl2[i] = anum2[i];
	anum2[i] = 0;
      }
    }

    for (i = 0; i < fitsstuff->nlt; i++) {
      fitsstuff->dki[i] = dnl2[i];
      totdk[i] = totdk[i] + fitsstuff->dki[i];
    }
    //printf("Done with flag != 2.\n");
  } // done with flag != 2
  
  /* --- this stuff is executed only if flag == 2 --- */
  /* --- flag = 2 means that recoils have stoped.
     --- we now decay everything left in each level 
     --- to the stopped peak --- */
  
  

  if (flag == 2) {
    //printf("Dealing with flag == 2.\n");
    dnt = 0;
    //printf("about to deal with dnt.\n");
    for ( i  = 0; i < 4; i++) {
      for ( j = 2; j < 10; j++) {
	dnt = dnt + anumi[i][j] + sfi[i][j];
      }
    }
    //printf("Done with dnt.\n");
    dnt = dnt = anumi[1][0] + anumi[1][1] + sfi[1][0] + sfi[1][1];
    if (fitsstuff->pop[0][1] < 0) {
      //printf("about to deal with dki.\n");
      for ( i = 0; i < fitsstuff->nlt; i++) {
	fitsstuff->dki[i] = sfi[0][0] + anum2[i] + dnt;
      }
      //printf("done with dki.\n");
    }
  }
  
}

/* GetStoppingPower
 *
 * Calculates the target stopping power
 * for a given energy by doing a linear interpolation
 * of input stopping power tables.
 */
void GetStoppingPower(float ei, int k, float *ef, int kstop, float *dedx)
{
  float ep,ep12,ep32;
  float eparg,flnfac;
  float top,bot;
  float depdrho,dedxn;
  int i,j;
  float dedxe,elo,ehi,slo,shi,rs;
  
  ep = fitsstuff->epfac[k] * ei;
  /* --- ncoice = 3 for not calculating nuclear stopping --- */
  if (stoppingpowerinfo->nchoice != 3) {
    if ((stoppingpowerinfo->nchoice == 1) || (stoppingpowerinfo->nchoice == 2)) {
      if (stoppingpowerinfo->nchoice == 1) {
	ep12 = sqrt(ep);
	ep32 = pow(ep,1.5);
	eparg = ep + 2.718282;
	flnfac = log(eparg);
	top = ep12 * flnfac;
	bot = 0.6 * ep12 - 0.6 * ep + 2 * ep32;
	depdrho = top/bot;
	dedxn = fitsstuff->facnuke[k] * depdrho;
      } else {
	if (ep != 0) {
	  flnfac = log(2 * ep);
	  dedxn = fitsstuff->facnuke[k] * flnfac / (2 * ep);
	  if (dedxn < 1) dedxn = 1;
	} else {
	  dedxn = 1;
	}
      }
    }
  }

  /* --- done with nuclear stopping --- */

  for ( i = 0; i < (stoppingpowerinfo->numcoarsestop - 1); i++) {
    if (k = 0) {
      if ( ei == stoppingpowerinfo->stopen1[i]) {
	dedxe = stoppingpowerinfo->stop1[i];
	goto loopdone;
      } else {
	if ((ei > stoppingpowerinfo->stopen1[i]) &&
	  (ei < stoppingpowerinfo->stopen1[i+1])) {
	  elo = stoppingpowerinfo->stopen1[i];
	  ehi = stoppingpowerinfo->stopen1[i+1];
	  slo = stoppingpowerinfo->stop1[i];
	  shi = stoppingpowerinfo->stop1[i+1];
	  dedxe = LinearInterPol(elo,ehi,slo,shi,ei);
	  goto loopdone;
	}
      }
    } else {
      if ( ei == stoppingpowerinfo->stopen2[i]) {
	dedxe = stoppingpowerinfo->stop2[i];
	goto loopdone;
      } else {
	if ((ei > stoppingpowerinfo->stopen2[i]) &&
	  (ei < stoppingpowerinfo->stopen2[i+1])) {
	  elo = stoppingpowerinfo->stopen2[i];
	  ehi = stoppingpowerinfo->stopen2[i+1];
	  slo = stoppingpowerinfo->stop2[i];
	  shi = stoppingpowerinfo->stop2[i+1];
	  dedxe = LinearInterPol(elo,ehi,slo,shi,ei);
	  goto loopdone;
	}
      }
    }
  }
 loopdone:
  *dedx = (float) stoppingpowerinfo->fe[k] * dedxe + stoppingpowerinfo->fn[k] * dedxn;
  rs = *dedx * stoppingpowerinfo->numrecoils[k] * stoppingpowerinfo->tarstopnorm[k];
  *ef = ei - rs * stoppingpowerinfo->tardensity;

}

/* ProjectileStop
 * 
 * Calculates the projectile stopping
 * power for a given beam energy and
 * computes the energy loss across the
 * target element, and returns the
 * final projectile energy
 */
float ProjectileStop(float beamenergy)
{
  float finbeamen;
  float dele,jlo,jhi,slo,shi,elo,ehi,rs;
  float s;
  
  dele = beamenergy - stoppingpowerinfo->minbeamen;
  jlo = (int) (dele / stoppingpowerinfo->debeamen) + 1;
  jhi = jlo + 1;
  slo = stoppingpowerinfo->projstop[(int)jlo];
  shi = stoppingpowerinfo->projstop[(int)jhi];
  elo = stoppingpowerinfo->minbeamen + (jlo - 1) * stoppingpowerinfo->debeamen;
  ehi = elo + stoppingpowerinfo->debeamen;
  s = LinearInterPol(elo,ehi,slo,shi,beamenergy);
  rs = stoppingpowerinfo->numrecoils[1] * s * stoppingpowerinfo->numcoarsestop;
  finbeamen = beamenergy - rs * stoppingpowerinfo->tardensity;
  return(finbeamen);
}

/* ChiSqr
 *
 * Calculates the Chi-square values
 * for each spectrum generated for each
 * lifetime of the level of interest, weighted
 * by the standard deviation squared, i. e.
 * sigma-square root of the background subtracted counts
 *
 * added 8-89 to fits.f: 
 * Returns the 10 lowest chi-square values, and the 
 * 10 best fitting theoretical spectra, bestsp, normalized to 
 * minimize chi-square.
 *
 * added 8-89 to fits.f:
 * optional chi-squared minimization f background if inut is -1.
 * The Fitted backgrounds are returned in bestbk.
 */
void ChiSqr()
{
  float np;
  int i,j,k;
  int jj,in;
  float sumxy,sumy2,sumx,sumy;
  float x,y;
  float sumt,snew[100];
  float npos[1000];
  int ic,icf[10],icmax;
  float chisq[1000];
  int ierr[100];
  float cnorm[1000];
  int kfit[2][1000];
  float xumxy,spnew,det;
  float back[1000];
  float csq;
  float sigsq;
  float bestc;

  printf("Beginning Chisqr.\n");

  np = (fitsstuff->pkhi - fitsstuff->pklo) / fitsstuff->dpeak;
  if ( np < 0) np = - np;
  if ( np > 10) np = 10;
  for ( i = 0; i < 1000; i++) {
    npos[i] = 0;
    kfit[0][i] = 0;
    kfit[1][i] = 1;
    chisq[i] = 0;
  }
  printf("1\n");
  ic = 0;
  for ( i = 0; i < np; i++) {
    fitsstuff->peakpos[i] = fitsstuff->pklo + (i - 1) * fitsstuff->dpeak;
    printf("2\n");
    for (j = 0; j < fitsstuff->nlt; j++) {
      chisq[ic] = 0;
      sumt = 0;
      npos[ic] = i;
      kfit[1][ic] = k;
      for (k = 0; k < 100; k++) {
	snew[k] = 0;
      }
      in = 0;
      sumxy = 0;
      sumy2 = 0;
      sumx = 0;
      sumy = 0;
      printf("3\n");
      if (fitsstuff->ibkg > 0) {
	printf("3-1\n");
	/* --- make a least squares fit to the norm only --- */
	for ( k = 0; k < fitsstuff->ichn; k++) {
	  jj = 600 + 10 * (k - fitsstuff->peakpos[i]);
	  snew[k] = fitsstuff->spect[i][jj];
	  if (fitsstuff->ispece[k] >= 0) {
	    in = in + 1;
	    x = fitsstuff->ispece[k] - fitsstuff->ibkg;
	    y = snew[k];
	    sumxy = sumxy + x * y;
	    sumy2 = sumy2 + pow(y,2);
	  }
	  printf("3-2\n");
	} // done k loop
	back[ic] = fitsstuff->ibkg;
	cnorm[ic] = 1;
	printf("3-3 ic = %d\n",ic);
	if (sumy2 != 0) cnorm[ic] = sumxy / sumy2;
      } // done with fitsstuff->ibkg greater than zero
      else {
	printf("4\n");
	/* --- fit both norm and bkg to minimize chi-square --- */
	for (k=0; k < fitsstuff->ichn; k++) {
	  jj = 600 + 10 * (k - fitsstuff->peakpos[i]);
	  snew[k] = fitsstuff->spect[j][jj];
	  if (fitsstuff->ispece[k] > 0) {
	    in = in + 1;
	    x = fitsstuff->ispece[k];
	    y = snew[k];
	    sumxy = sumxy + x * y;
	    sumy2 = sumy2 + pow(y,2);
	    sumx = sumx + x;
	    sumy = sumy + y;
	  }
	} // done with k loop
	det = in * sumy2 - pow(sumy,2);
	printf("5\n");
	cnorm[ic] = 1;
	back[ic] = 0;
	if (det != 0) {
	  cnorm[ic] = (in * xumxy - sumx * sumy) / det;
	  back[ic] = (sumy2 * sumx - sumy * sumxy) / det;
	}
      } // done with fitsstuff->ibkg < 0
      printf("6\n");
      for ( k = 0; k < fitsstuff->ichn; k++) {
	//printf("ichn = %d\n",(int)fitsstuff->ichn);
	spnew = cnorm[ic] * snew[k];
	if (fitsstuff->ispece[k] > 0) {
	  sigsq = (float) fitsstuff->ispece[k];
	  if (sigsq > 0) {
	    csq = spnew - (sigsq - back[ic]);
	    chisq[ic] = chisq[ic] + pow(csq,2) / (in * sigsq);
	  }
	}
      } // done with k loop
      ic++;
      printf("7\n");
    } // end j loop
    printf("Finished j loop\n");
  } // end i loop
  printf("finished i loop\n");

  icmax = ic - 1;
  printf("8\n");
  if ( icmax < 10) {
    for ( i = 0; i < 10; i++) {
      icf[i] = i;
    }
  } else {
    printf("9\n");
    for (i = 0; i < icmax; i++) {
      if ( i < 10) icf[i] = i;
      for ( j = 0; j < 10; j++ ) {
	if ( chisq[i] < chisq[icf[j]]) {
	  icf[j] = i;
	}
      }
    }
  }
  printf("10\n");
  for ( i < 0; i < 10; i++) {
    k = kfit[1][icf[i]];
    fitsstuff->nbest[i] = npos[icf[i]];
    fitsstuff->bestt1[i] = fitsstuff->taul[k];
    fitsstuff->bestt2[i] = 0;
    fitsstuff->bestchi[i] = chisq[icf[i]];
    bestc = cnorm[icf[i]];
    fitsstuff->bestbk[i] = back[icf[i]];
    for ( j = 0; j < 1200; j++) {
      fitsstuff->bestsp[i][j] = bestc * fitsstuff->spect[i][j];
    }
  }
  printf("11\n");
}
